from django.db.models.signals import post_save, m2m_changed
from django.dispatch import receiver
from django.db.models import F

from bcmr.models import Token, Registry


@receiver(post_save, sender=Token)
def token_post_save(sender, instance=None, created=False, **kwargs):
    if not created:
        token_registries = instance.registry_set.all()
        token_registries.update(patch=F('patch') + 1)


@receiver(m2m_changed, sender=Registry.tokens.through)
def registry_tokens_changed(sender, **kwargs):
    action = kwargs['action']
    instance = kwargs['instance']

    if action == 'pre_remove':
        instance.major += 1
        instance.save()
    elif action == 'pre_add':
        instance.minor += 1
        instance.save()



from django.core.exceptions import ImproperlyConfigured
from django.utils import timezone
from django.conf import settings

from rest_framework.response import Response
from rest_framework import viewsets, mixins

from bcmr.serializers import *
from bcmr.models import *


class TokenViewSet(
    viewsets.GenericViewSet,
    mixins.CreateModelMixin,
    mixins.UpdateModelMixin,
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin
):
    queryset = Token.objects.all()
    serializer_class = CashTokenSerializer


class RegistryViewSet(
    viewsets.GenericViewSet,
    mixins.CreateModelMixin,
    mixins.UpdateModelMixin,
    mixins.ListModelMixin,
    mixins.RetrieveModelMixin
):
    queryset = Registry.objects.all()
    serializer_class = EmptySerializer

    serializer_classes = {
        'create': RegistrySerializer,
        'list': BcmrRegistrySerializer,
        'update': RegistrySerializer,
        'partial_update': RegistrySerializer,
        'retrieve': BcmrRegistrySerializer
    }

    def update(self, request, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        instance.latest_revision = timezone.now()
        instance.save()

        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        bcmr = BcmrRegistrySerializer(instance)
        return Response(bcmr.data)

    def get_serializer_class(self):
        if not isinstance(self.serializer_classes, dict):
            raise ImproperlyConfigured("serializer_classes should be a dict mapping.")

        if self.action in self.serializer_classes.keys():
            return self.serializer_classes[self.action]
        return super().get_serializer_class()



from django.utils import timezone
from django.db import models


class Token(models.Model):
    class Status(models.TextChoices):
        ACTIVE = 'active'
        INACTIVE = 'inactive'
        BURNED = 'burned'

    category = models.CharField(max_length=255, primary_key=True, unique=True)
    name = models.CharField(max_length=255)
    description = models.CharField(max_length=255, null=True, blank=True, default='')
    symbol = models.CharField(max_length=100)
    decimals = models.PositiveIntegerField(default=0)
    icon = models.ImageField(null=True, blank=True)
    date_created = models.DateTimeField(default=timezone.now)
    status = models.CharField(
        max_length=10,
        choices=Status.choices,
        default=Status.ACTIVE
    )

    class Meta:
        ordering = ('name', )


class Registry(models.Model):
    major = models.PositiveIntegerField(default=0) # incremented when an identity is removed
    minor = models.PositiveIntegerField(default=0) # incremented when an identity is added
    patch = models.PositiveIntegerField(default=0) # incremented when an identity is modified
    name = models.CharField(max_length=255)
    description = models.CharField(max_length=255, null=True, blank=True)
    date_created = models.DateTimeField(default=timezone.now)
    latest_revision = models.DateTimeField(default=timezone.now)
    tokens = models.ManyToManyField(Token)

    class Meta:
        verbose_name_plural = 'Registries'



from rest_framework import routers

from bcmr import views


app_name = "bcmr"

router = routers.DefaultRouter()

router.register("registries", views.RegistryViewSet)
router.register("tokens", views.TokenViewSet)

urlpatterns = router.urls



from django.contrib import admin

from bcmr.models import *


admin.site.site_header = 'Django BCMR Admin'


class TokenAdmin(admin.ModelAdmin):
    list_display = [
        'category',
        'name',
        'symbol',
        'decimals',
        'status',
        'date_created',
    ]

class RegistryAdmin(admin.ModelAdmin):
    list_display = [
        'name',
        'description',
        'date_created',
        'latest_revision',
    ]


admin.site.register(Token, TokenAdmin)
admin.site.register(Registry, RegistryAdmin)
